/*******************************************************************
 *
 *	File:		i_audio.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2003, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/


#include "audio.h"
#include "polledas.h"

static CGameAudio* sGameAudio= NULL;

extern "C"
{
#include "doomdef.h"
#include "i_system.h"
#include "m_argv.h"
#include "st_stuff.h"
#include "i_video.h"
#include "z_zone.h"
#include "m_misc.h"


#include "doomstat.h"
#include "s_sound.h"
consvar_t cdUpdate  = {"cd_update","1",CV_SAVE};
#include "i_sound.h"
extern CV_PossibleValue_t soundvolume_cons_t[];
consvar_t   cd_volume = {"cd_volume","31",CV_SAVE,soundvolume_cons_t};

extern consvar_t cv_vibrate;

void VibrateWhenInPain(int);

}		// extern "C"


/*********************************************************
 *
 * Vibration stuff
 *
 *********************************************************/

#include "vibration.h"


class TVibrationObserver : public SonyEricsson::MVibrationObserver
{
public:
	virtual void VibrationResponse(SonyEricsson::CVibration* /*aVibration*/, TInt /*aErr*/){iVibratorOn=EFalse;}
	TBool iVibratorOn;
};

static TVibrationObserver			sVibrationObserver;
static SonyEricsson::CVibration*	sVibrator = NULL;

void VibrateWhenInPain(int aPainLevel)
{
#ifndef __WINS__
	if (sVibrator && (cv_vibrate.value == 1) && (!sVibrationObserver.iVibratorOn))
	{
//		sVibrator->VibrationOff();
		TInt noOfTimes = aPainLevel / 10;
		if (noOfTimes < 2 )
			noOfTimes = 2;
		sVibrationObserver.iVibratorOn = ETrue;
		sVibrator->VibrationOn(2, 2, noOfTimes);
	}
#endif
}



/**************************************
 *
 * AUDIO
 *
 **************************************/

// Number of sound channels used as set in the config file
extern  consvar_t cv_numChannels;

// Flags for the -nosound and -nomusic options
extern boolean nosound;
extern boolean nomusic;


TInt		gHardwareVolume = 31;


/**********************************
 *
 * Mixer
 *
 **********************************/

#define FRAMERATE				17
#define BUFMUL                  2		// 2 for two channels.
#define SAMPLERATE              11025   // Hz
#define SAMPLESIZE              2       // 16bit
#define SAMPLECOUNT             (SAMPLERATE / FRAMERATE)
#define MIXBUFFERSIZE           (SAMPLECOUNT*BUFMUL)

class TMixer
{
public:
	void	Init();
	static TMixer iMixer;

	TInt16	iMixBuffer[MIXBUFFERSIZE];
	TInt	iVolume;
	TInt	MixBufferLengthInBytes();
	TBool	iIsStereo;
	TInt	iVolumeTable[128*256];
	TInt	iStepTable[256];
};


TMixer TMixer::iMixer;

void TMixer::Init()
{
	TInt*	steptablemid = &iStepTable[128];
	for (TInt s=-128 ; s<128 ; s++)
		steptablemid[s] = (int)(pow(2.0, (s/64.0))*65536.0);

	for (TInt i=0 ; i<128 ; i++)
	{
		for (TInt j=0 ; j<256 ; j++)
		{
			if (iIsStereo)
				iVolumeTable[i*256+j] = (i*(j-128)*256)/127;
			else
				iVolumeTable[i*256+j] = (i*(j-128)*256)/127*4;
		}
	}
}


CPolledActiveScheduler*		sTheScheduler;
 


/**************************************
 *
 * Channel
 *
 **************************************/

#define NUM_CHANNELS			16

class TChannel
{
public:
	TBool	IsActive(){return (iData!=NULL);}
	TInt	Allocate(
				TInt		aSfxId, 
				TInt		aVolume,
				TInt		aSeperation,
				TInt		aPitch,
				TInt		aPriority);
	void	FreeChannel();
	TInt	Handle() {return iHandle;}
	void	Sample(TInt& aSampleLeft, TInt& aSampleRight);
	void	UpdateData();

	static TChannel* FindFreeChannel(TInt aSfxId, TInt aPriority);
	static TChannel* FindChannelByHandle(TInt aHandle);

	static TChannel iChannels[NUM_CHANNELS];

protected:
	TUint8*	iData;
	TInt	iDataLength;
	TInt	iHandle;
	TInt	iPriority;
	TUint	iStartTic;
	TInt	iStep;
	TInt	iStepRemainder;
	TInt*	iVolumeLookupLeft;
	TInt*	iVolumeLookupRight;

	static TInt iUniqueId;
};

TInt TChannel::iUniqueId = 1;
TChannel TChannel::iChannels[NUM_CHANNELS];


inline void TChannel::Sample(TInt& aSampleLeft, TInt& aSampleRight)
{
	TInt sample = *iData; 

	aSampleLeft = iVolumeLookupLeft[sample];
	aSampleRight = iVolumeLookupRight[sample];

	UpdateData();
}

inline void	TChannel::UpdateData()
{
    iStepRemainder += iStep;
	TInt step = iStepRemainder >> 16;
    iStepRemainder &= 0xffff;

    iData += step;
    iDataLength -= step;

	if (iDataLength <= 0)
		FreeChannel();
}


TChannel* TChannel::FindFreeChannel(TInt /*aSfxId*/, TInt aPriority)
{
	TChannel*	channel = &iChannels[0];
	TChannel*	freeChannel = NULL;
	for (TInt i=0 ; i<cv_numChannels.value ; i++, channel++)
	{
		if (!channel->IsActive())
			return channel;
		else if (channel->iPriority <= aPriority)
		{
			if (!freeChannel)
				freeChannel = channel;
			else
			{
				if (channel->iPriority < freeChannel->iPriority)
					freeChannel = channel;
				else if ((channel->iPriority == freeChannel->iPriority) && (channel->iStartTic <freeChannel->iStartTic))
					freeChannel = channel;
			}
		}
	}
	return freeChannel;
}


TChannel* TChannel::FindChannelByHandle(TInt aHandle)
{
	TChannel*	channel = &iChannels[0];
	for (TInt i=0 ; i<cv_numChannels.value ; i++, channel++)
	{
		if (channel->iHandle == aHandle)
			return channel;
	}
	return NULL;
}


TInt TChannel::Allocate(
	TInt		aSfxId, 
	TInt		aVolume,
	TInt		aSeperation,
	TInt		aPitch,
	TInt		aPriority)
{
	iHandle = TChannel::iUniqueId++;

	iPriority = aPriority;
	iStartTic = gametic;

	iData = ((TUint8*) S_sfx[aSfxId].data)+8;
	iDataLength = *((int*)S_sfx[aSfxId].data);

	iStep = TMixer::iMixer.iStepTable[aPitch];
	iStepRemainder = 0;

    TInt leftVol = aVolume - ((aVolume * aSeperation * aSeperation) >> 16);
    aSeperation = aSeperation - 257;
    TInt rightVol = aVolume - ((aVolume * aSeperation * aSeperation) >> 16);

	if (leftVol<0) leftVol = 0;
	if (leftVol>127) leftVol = 127;
	if (rightVol<0) rightVol = 0;
	if (rightVol>127) rightVol = 127;

	leftVol = (leftVol * TMixer::iMixer.iVolume) / 31;
	rightVol = (rightVol * TMixer::iMixer.iVolume) / 31;

	iVolumeLookupLeft = &TMixer::iMixer.iVolumeTable[leftVol * 256];
	iVolumeLookupRight = &TMixer::iMixer.iVolumeTable[rightVol * 256];

	return iHandle;
}



void TChannel::FreeChannel()
{
//	RDebug::Print(_L("Stop channel, handle=%d"), aChannel->iHandle);
	iHandle = -1;
	iData = NULL;
	iDataLength = 0;
}





void I_SetSfxVolume(TInt aVolume)
{
	TMixer::iMixer.iVolume = aVolume;
}


TInt I_StartSound(
	TInt		aSfxId,
	TInt		aVolume,
	TInt		aSep,
	TInt		aPitch,
	TInt		aPriority)
{
	aVolume >>= 1;

//	RDebug::Print(_L("StartSound: %d"), aSfxId);

	TChannel*	channel = NULL;
	TInt	handle = -1;

	if ( aSfxId == sfx_sawup
			|| aSfxId == sfx_sawidl
			|| aSfxId == sfx_sawful
			|| aSfxId == sfx_sawhit
			|| aSfxId == sfx_stnmov
			|| aSfxId == sfx_pistol  
			)
	{
		TChannel*	channel = &TChannel::iChannels[0];
		for (TInt i=0 ; i<cv_numChannels.value ; i++, channel++)
			if (channel->Handle() == aSfxId)
				channel->FreeChannel();
	}


	channel = TChannel::FindFreeChannel(aSfxId, aPriority);
	if (channel)
	{
		handle = channel->Allocate(aSfxId, aVolume, aSep, aPitch, aPriority);
//		RDebug::Print(_L("Start channel, aSfxId=%d, handle=%d, step=%d"), aSfxId, channel->iHandle, channel->iStep);
	}

	return handle;
}


void I_StopSound (TInt aHandle)
{
	TChannel* channel = TChannel::FindChannelByHandle(aHandle);
	if (channel)
		channel->FreeChannel();
}


int I_SoundIsPlaying(TInt aHandle)
{
	TChannel* channel = TChannel::FindChannelByHandle(aHandle);
	return (channel) ? 1 : 0;
}



static TInt sActive = 0;

void I_UpdateSound( void )
{
	TInt16*		mixData = &TMixer::iMixer.iMixBuffer[0];
	TInt		size = SAMPLECOUNT;

	sActive = 0;
	while (size--)
	{
		TInt	dl = 0;
		TInt	dr = 0;

		TChannel*	channel = &TChannel::iChannels[0];
		for (TInt i=0 ; i<cv_numChannels.value ; i++, channel++)
		{
			if (channel->IsActive())
			{
				sActive = 1;

				TInt sampleLeft, sampleRight;
				channel->Sample(sampleLeft, sampleRight);
				dl += sampleLeft;
				dr += sampleRight;
			}
		}
				
		if (dl > 0x7fff)
			dl = 0x7fff;
		else if (dl < -0x8000)
			dl = -0x8000;

		if (dr > 0x7fff)
			dr = 0x7fff;
		else if (dr < -0x8000)
			dr = -0x8000;

		if (TMixer::iMixer.iIsStereo)
		{
			*mixData++  = (TInt16) dl;
			*mixData++  = (TInt16) dr;
		}
		else
		{
	        TUint16	sdl = dl ^ 0xfff8000;
	        TUint16	sdr = dr ^ 0xfff8000;
			*mixData++ = (((sdr + sdl) / 2) >> 8);

		}
	}
}


void
I_SubmitSound(void)
{
	sTheScheduler->Schedule();
	if (sActive)
		sGameAudio->Write((TInt16*)TMixer::iMixer.iMixBuffer, SAMPLECOUNT);
}



void
I_UpdateSoundParams
( int   /*handle*/,
  int   /*vol*/,
  int   /*sep*/,
  int   /*pitch*/)
{
}




void I_ShutdownSound(void)
{    
	delete sVibrator;
	sVibrator = NULL;
}




void
I_StartupSound()
{ 
	sTheScheduler = CPolledActiveScheduler::NewL();

//	RDebug::Print(_L("I_StartupSound: create vibrator"));
//	RDebug::Print(_L("I_StartupSound: create vibrator"));
#ifndef __WINS__
	TRAPD(err, sVibrator = SonyEricsson::CVibration::NewL());
	sVibrationObserver.iVibratorOn = EFalse;
#endif
//	RDebug::Print(_L("I_StartupSound: done creating vibrator"));

	if (nosound)
		return;

	TMixer::iMixer.iIsStereo = ETrue;
	TMixer::iMixer.Init();

	sGameAudio = CGameAudio::NewL(TMixer::iMixer.iIsStereo, SAMPLECOUNT);
	memset(TMixer::iMixer.iMixBuffer, 0, sizeof(TMixer::iMixer.iMixBuffer));

}



//
// This function loads the sound data from the WAD lump,
//  for single sound.
//
void* I_GetSfx (sfxinfo_t*  sfx)
{
    byte*               dssfx;
    int                 size,i;
    unsigned char*      paddedsfx;
    int                 paddedsize;

    if (sfx->lumpnum<0)
    {
        sfx->lumpnum = S_GetSfxLumpNum (sfx);
    }

    size = W_LumpLength (sfx->lumpnum);

    dssfx = (byte*) W_CacheLumpNum (sfx->lumpnum, PU_STATIC);
    paddedsize = ((size-8 + (SAMPLECOUNT-1)) / SAMPLECOUNT) * SAMPLECOUNT;
    paddedsfx = (unsigned char*)Z_Malloc( paddedsize+8, PU_STATIC, 0 );
    memcpy(  paddedsfx, dssfx, size );
    for (i=size ; i<paddedsize+8 ; i++)
        paddedsfx[i] = 128;

    // Remove the cached lump.
    Z_Free( dssfx );
    *((int*)paddedsfx) = paddedsize;
    return (void*)(paddedsfx);

}

// dummy for now 19990220 by Kin
void I_FreeSfx(sfxinfo_t* /*sfx*/) 
{
}

//
// Retrieve the raw data lump index
//  for a given SFX name.
//
int I_GetSfxLumpNum(sfxinfo_t* sfx)
{
    char namebuf[9];
    sprintf(namebuf, "ds%s", sfx->name);
    return W_GetNumForName(namebuf);
}
