/*******************************************************
 *
 *  File: ll_video.c
 *
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2001, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************/

#include "ll_video.h"
#include "font_8x8.h"

#define NULL ((void*) 0)


#define MK_DWORD_PIXELS_2BPP(c) (( (c) << 30) | ( (c) << 28) | ( (c) << 26) | ( (c) << 24) | ( (c) << 22) | \
								 ( (c) << 20) | ( (c) << 18) | ( (c) << 16) | ( (c) << 14) | ( (c) << 12) | \
								 ( (c) << 10) | ( (c) <<  8) | ( (c) <<  6) | ( (c) <<  4) | ( (c) <<  2) | (c) )
#define MK_DWORD_PIXELS_4BPP(c) (( (c) << 28) | ( (c) << 24) | ( (c) << 20) | ( (c) << 16) | ( (c) << 12) | \
								 ( (c) <<  8) | ( (c) <<  4) | (c) )
#define MK_DWORD_PIXELS_8BPP(c) (( (c) << 24) | ( (c) << 16) | ( (c) << 8) | (c) )


#ifdef _DEBUG
void video_write(int offset, int length)
{
	unsigned*	video = (unsigned*) 0xc0000200 + (640*offset);
	int			i;
	for(i=0 ; i<length ; i++)
		*video++ = 0;
}
#endif

/*******************************************
 *
 * static global data
 *
 *******************************************/


static unsigned		s_char_xpos;			// current cursor x
static unsigned		s_char_ypos;			// current cursor y
static unsigned		s_char_width;			// number of characters per line
static unsigned		s_char_height;			// number of lines

static unsigned*	s_video_base;			// start address of the frame buffer
static unsigned		s_pixel_width;
static unsigned		s_pixel_height;

static unsigned		s_font_width;
static unsigned		s_font_height;

static unsigned		s_fg_colour;
static unsigned		s_bg_colour;

/*******************************************
 *
 * Video depth dependent routines
 *
 *******************************************/

typedef struct _video_if
{
	int			bpp;
	unsigned	scanline_dword_size;
	unsigned	bg_dword_pixels;		// DWORD worth of background pixels in the given video mode
	unsigned	fg_dword_pixels;		// DWORD worth of foreground pixels in the given video mode
	
		/*
		 * Can't use a jump table because these won't be GOT relative...
		 */
	// void 		(*setup)(struct _video_if* video_if);
	// void 		(*print_fontchar)(struct _video_if* video_if, const unsigned char* fontchar, int xpos, int ypos);
} video_if_t;


/*******************************************
 *
 * 2 bpp video
 *
 *******************************************/

static void video_2bpp_setup(video_if_t* video_if)
{
	video_if->scanline_dword_size = s_pixel_width >> 4;
	video_if->bg_dword_pixels = MK_DWORD_PIXELS_2BPP(s_bg_colour & 0x3);
	video_if->fg_dword_pixels = MK_DWORD_PIXELS_2BPP(s_fg_colour & 0x3);
}

static void video_2bpp_print_fontchar(video_if_t* video_if, const unsigned char* fontchar, int xpos, int ypos)
{
	unsigned char*	fb_line = ((unsigned char*)s_video_base)
					+ (ypos * video_if->scanline_dword_size * s_font_height * 4)
					+ (((xpos+1) * s_font_width) >> 2);
	unsigned char*	fb;
	unsigned		x, y, bitmap;
	unsigned		count;
	unsigned char	byte;
	
	for(y=0 ; y<s_font_height ; y++)
	{
		bitmap = *fontchar++;
		fb = (unsigned char*) fb_line;
		count=0;
		byte =0;
		for(x=0 ; x<s_font_width ; x++)
		{
			if (bitmap & 1)
				byte |= (s_fg_colour & 0x3);
			else
				byte |= (s_bg_colour & 0x3);
			if (++count == 4)	
			{
				fb--;
				*(fb) = (unsigned char) byte;
				count = 0;
				byte = 0;
			}
			bitmap>>=1;
			byte<<=2;
		}
		fb_line += video_if->scanline_dword_size * 4;
	}
}


/*******************************************
 *
 * 4 bpp video
 *
 *******************************************/

static void video_4bpp_setup(video_if_t* video_if)
{
	video_if->scanline_dword_size = s_pixel_width >> 3;
	video_if->bg_dword_pixels = MK_DWORD_PIXELS_4BPP(s_bg_colour & 0xf);
	video_if->fg_dword_pixels = MK_DWORD_PIXELS_4BPP(s_fg_colour & 0xf);
}

static void video_4bpp_print_fontchar(video_if_t* video_if, const unsigned char* fontchar, int xpos, int ypos)
{
	unsigned char*	fb_line = ((unsigned char*)s_video_base)
					+ (ypos * video_if->scanline_dword_size * s_font_height * 4)
					+ (((xpos+1) * s_font_width) >> 1);
	unsigned char*	fb;
	unsigned		x, y, bitmap;
	unsigned		count;
	unsigned char	byte;
	
	for(y=0 ; y<s_font_height ; y++)
	{
		bitmap = *fontchar++;
		fb = (unsigned char*) fb_line;
		count=0;
		byte =0;
		for(x=0 ; x<s_font_width ; x++)
		{
			if (bitmap & 1)
				byte |= (s_fg_colour & 0xf);
			else
				byte |= (s_bg_colour & 0xf);
			if (++count == 2)	
			{
				fb--;
				*(fb) = (unsigned char) byte;
				count = 0;
				byte = 0;
			}
			bitmap>>=1;
			byte<<=4;
		}
		fb_line += video_if->scanline_dword_size * 4;
	}
}


/*******************************************
 *
 * 8 bpp video
 *
 *******************************************/

static void video_8bpp_setup(video_if_t* video_if)
{
	video_if->scanline_dword_size = s_pixel_width >> 2;
	video_if->bg_dword_pixels = MK_DWORD_PIXELS_8BPP(s_bg_colour & 0xff);
	video_if->fg_dword_pixels = MK_DWORD_PIXELS_8BPP(s_fg_colour & 0xff);
}


static void video_8bpp_print_fontchar(video_if_t* video_if, const unsigned char* fontchar, int xpos, int ypos)
{
	unsigned*		fb_line = s_video_base + (ypos * video_if->scanline_dword_size * s_font_height) + (( (xpos+1) * s_font_width) >> 2);
	unsigned char*	fb;
	unsigned		x, y, bitmap;
	
	for(y=0 ; y<s_font_height ; y++)
	{
		bitmap = *fontchar++;
		fb = (unsigned char*) fb_line;
		for(x=0 ; x<s_font_width ; x++)
		{
			fb--;
			if (bitmap & 1)
				*(fb) = (unsigned char) s_fg_colour;
			bitmap>>=1;
		}
		fb_line += video_if->scanline_dword_size;
	}
}

/*******************************************
 *
 * 12 bpp video
 *
 *******************************************/

static void video_12bpp_setup(video_if_t* video_if)
{
	video_if->scanline_dword_size = s_pixel_width >> 2;
	video_if->bg_dword_pixels = MK_DWORD_PIXELS_8BPP(s_bg_colour & 0xff);
	video_if->fg_dword_pixels = MK_DWORD_PIXELS_8BPP(s_fg_colour & 0xff);
}


static void video_12bpp_print_fontchar(video_if_t* video_if, const unsigned char* fontchar, int xpos, int ypos)
{
	unsigned*		fb_line = s_video_base + (ypos * video_if->scanline_dword_size * s_font_height) + (( (xpos+1) * s_font_width) >> 2);
	unsigned char*	fb;
	unsigned		x, y, bitmap;
	
	for(y=0 ; y<s_font_height ; y++)
	{
		bitmap = *fontchar++;
		fb = (unsigned char*) fb_line;
		for(x=0 ; x<s_font_width ; x++)
		{
			fb--;
			if (bitmap & 1)
				*(fb) = (unsigned char) s_fg_colour;
			bitmap>>=1;
		}
		fb_line += video_if->scanline_dword_size;
	}
}


/*******************************************
 *
 * Video config/selection
 *
 *******************************************/

static video_if_t	videos[] =
{
	{
		2, 0, 0, 0
	},
	{
		4, 0, 0, 0
	},
	{
		8, 0, 0, 0
	},
	{
		12, 0, 0, 0
	},
	{
		0, 0, 0, 0
	}
};

static video_if_t*		s_the_video;


static video_if_t* get_video(int bpp)
{
	video_if_t*		video = &videos[0];
	
	while(video->bpp > 0)
	{
		if (video->bpp == bpp)
			return video;
		video++;
	}
	
	return NULL;
}


static void video_xbpp_setup(video_if_t* video_if)
{
	switch(video_if->bpp)
	{
		case 2:
			video_2bpp_setup(video_if);
			break;
		case 4:
			video_4bpp_setup(video_if);
			break;
		case 8:
			video_8bpp_setup(video_if);
			break;
		case 12:
			video_12bpp_setup(video_if);
			break;
		default:
			break;
	}
}

static void video_xbpp_print_fontchar(video_if_t* video_if, const unsigned char* fontchar, int xpos, int ypos)
{
	switch(video_if->bpp)
	{
		case 2:
			video_2bpp_print_fontchar(video_if, fontchar, xpos, ypos);
			break;
		case 4:
			video_4bpp_print_fontchar(video_if, fontchar, xpos, ypos);
			break;
		case 8:
			video_8bpp_print_fontchar(video_if, fontchar, xpos, ypos);
			break;
		case 12:
//			video_12bpp_print_fontchar(video_if, fontchar, xpos, ypos);
			break;
		default:
			break;
	}
}


/*******************************************
 *
 * Video depth independent routines
 *
 *******************************************/

/*
 * On the series 7:
 * Blue: 220
 * Red: 5
 * Yellow: 250
 */

static void video_printch_pos(char c, int xpos, int ypos)
{
	unsigned const char*  fontchar = &fontdata_8x8[c * s_font_height];

	video_xbpp_print_fontchar(s_the_video, fontchar, xpos, ypos);	
}


static void video_scroll_up(void)
{
	
	unsigned*	src = s_video_base + (s_font_height * s_the_video->scanline_dword_size);
	unsigned*	dst = s_video_base;
	unsigned	num_lines = s_pixel_height - s_font_height;
	unsigned	count;

	count = num_lines * s_the_video->scanline_dword_size;
	while(count--)
		*dst++ = *src++;
		
	count = s_font_height* s_the_video->scanline_dword_size;
	
	while(count--)
		*dst++ = s_the_video->bg_dword_pixels;
}


static void video_linefeed(void)
{
	s_char_xpos=0;
	if (++s_char_ypos == s_char_height)
		video_scroll_up();
}



static void video_printch(char c)
{
	if (c == '\n')
	{
		video_linefeed();
		return;
	}
		
	video_printch_pos(c, s_char_xpos, s_char_ypos);
	
	if (++s_char_xpos == s_char_width)
		video_linefeed();
}


/*******************************************
 *
 * Interface routines
 *
 *******************************************/

void video_setup(unsigned* video_base, int width, int height, int bpp, unsigned fg_colour, unsigned bg_colour)
{
	s_video_base = video_base;
	s_font_width = 8;
	s_font_height = 8;

	s_pixel_width = width;
	s_pixel_height = height;
	s_char_width = width / s_font_width;
	s_char_height = height / s_font_height;
	s_fg_colour = fg_colour;
	s_bg_colour = bg_colour;
	s_char_xpos = s_char_ypos = 0;
	
	s_the_video = get_video(bpp);
	
	if (s_the_video)
		video_xbpp_setup(s_the_video);
}



void video_clear()
{
	unsigned*	dst;
	unsigned	count;
	
	if (s_the_video)
	{
		dst = s_video_base;
    	count = s_pixel_height * s_the_video->scanline_dword_size;
        	
		while(count--)
			*dst++ = s_the_video->bg_dword_pixels;
		
		s_char_xpos = s_char_ypos = 0;
	}
}


void video_puts(const char* str)
{
	if (s_the_video)
	{
	    while(*str)
			video_printch(*str++);
	}
}


void video_puts_pos(const char* str, int x, int y)
{
	s_char_xpos = x;
	s_char_ypos = y;
	video_puts(str);
}

void video_clear_area(int x, int y, int w, int h)
{
	unsigned*		fb_line = s_video_base + (y * s_the_video->scanline_dword_size * s_font_height) + ((x * s_font_width) >> 2);
	unsigned char*	dst;
	unsigned		lines, pixels;

	if (s_the_video)
	{
    	lines = s_font_height * h;
		while(lines--)
		{
	    	pixels = (s_font_width * w * 8) / s_the_video->bpp;
			dst = (unsigned char*)fb_line;
			while(pixels--)
				*dst++ = (unsigned char) s_the_video->bg_dword_pixels;
			fb_line += s_the_video->scanline_dword_size;
		}
	}
}
